const { args: contextArgs, normalizeGenerator } = require('./helpers')
const Lines = require('../lines')
const sortJson = require('sort-json')
const helpers = require('./helpers')

class Generator {
  // constructor ({ match, apply, uuid, index, km, priority, notes }) {
  constructor (generator) {
    generator = normalizeGenerator(generator)
    const { match, apply, uuid, index, km, priority, where, notes, debug } = generator
    this.match = match
    this._apply = apply
    this.uuid = uuid
    this.index = index
    this.km = km
    this.priority = priority || 0
    this.notes = notes
    this.where = where
    this.callId = debug
  }

  getAPI (config) {
    if (config && config.getAPI) {
      return config.getAPI(this.uuid)
    }
  }

  getAPIs (config) {
    if (config && config._api && config._api.multiApi) {
      return config._api.apis
    }
  }

  toLabel () {
    const where = this.where ? `where: "${this.where}"` : ''
    if (!this.km) {
      return where
    }
    return `KM '${this.km}' ordinal: ${this.index} ${where}`
  }

  toString () {
    return `Generator(${this.match}, ${this._apply})`
  }

  matches (baseArgs, objects, context, hierarchy, config, options = {}) {
    if (objects && objects.namespaced) {
      objects = objects.namespaced[this.uuid]
    }
    // return this.match({ ...args, args: contextArgs(context, hierarchy), objects, global: objects, hierarchy, context, api: this.getAPI(config) })
    const callId = baseArgs.calls.current()
    const moreArgs = { 
            uuid: this.uuid, 
            args: contextArgs(context, hierarchy), 
            objects, 
            global: objects, 
            // hierarchy, 
            context, 
            callId,
            api: this.getAPI(config),
            apis: this.getAPIs(config) 
          }
    const args = Object.assign({}, baseArgs, moreArgs)
    // return this.match(args)
    const matches = this.match(args)
    if ((matches && (options.debug || {}).match)
        ||
        callId == this.callId) {
      debugger; // next line is the matcher
      this.match(args)
    }
    return matches
  }

  apply (baseArgs, objects, g, gs, context, hierarchy, config, response, log, options = {}) {
    if (!log) {
      throw 'generators.apply argument log is required'
    }

    // this.getAPI(config)
    let n = (id) => id
    if (config && 'nsToString' in config) {
      n = (id) => config.nsToString(id)
    }
    if (objects && objects.namespaced) {
      objects = objects.namespaced[this.uuid]
    }
    const km = (name) => config.getConfig(name)
    const callId = baseArgs.calls.current()
    const moreArgs = { 
            callId, 
            uuid: this.uuid, 
            // km, 
            args: contextArgs(context, hierarchy), 
            objects, 
            log, 
            global: 
            objects, 
            g, 
            n, 
            hierarchy, 
            context, 
            uuid: this.uuid, 
            gs, 
            config, 
            response, 
            api: this.getAPI(config), 
            apis: this.getAPIs(config) 
          }
    const args = Object.assign({}, baseArgs, moreArgs)
    // if (this.callId) {
    // greg
    /*
    if (callId == 'call11' && this.callId) {
      debugger;
    }
    */
    if ((options.debug || {}).apply
        ||
        callId == this.callId) {
      debugger
    }
    return this._apply(args)
  }
}

class Generators {
  constructor (generators, logs = []) {
    generators = (generators || []).map((generator) => {
      if (generator instanceof Generator) {
        return generator
      } else {
        return new Generator(generator)
      }
    })
    this.defaults = []
    const priorityToGenerators = {}
    const add = (priority, value) => {
      if (!priorityToGenerators[priority]) {
        priorityToGenerators[priority] = []
      }
      priorityToGenerators[priority].push(value)
    }
    for (const generator of generators) {
      const priority = generator.priority
      add(priority, generator)
    }
    this.generators = []
    for (const key of Object.keys(priorityToGenerators).sort()) {
      this.generators = this.generators.concat(priorityToGenerators[key])
    }
    // this.generators = generators
    this.logs = logs
  };

  // assumed - properties added to contexts before the generators are called. For setting paraphrase property
  apply (args, contexts, assumed = {}, options = {}) {
    const config = args.config
    const objects = args.objects
    const hierarchy = args.hierarchy
    const response = args.response
    if (Array.isArray(contexts)) {
      // no-op
    } else if (typeof contexts === 'object') {
      contexts = [contexts]
    } else {
      return String(contexts)
    }

    const contextsPrime = []
    for (const icontext in contexts) {
      let context = contexts[icontext]
      context = Object.assign({}, context, assumed)
      // let context_prime = JSON.stringify(context);
      let contextPrime = context
      let applied = false
      const stack = args.calls.push()
      for (let igenerator = 0; igenerator < this.generators.length; ++igenerator) {
        const generator = this.generators[igenerator]
        if (generator.matches(args, objects, context, hierarchy, config, options)) {
          const g = (context, options) => {
            const r = this.apply(args, context, Object.assign({}, ((options||{}).assumed || {}), assumed), options)
            if (Array.isArray(r)) {
              return r.join(' ')
            } else {
              return r
            }
          }
          const log = (message) => { this.logs.push(message) }
          // this.logs.push(`Generators: applied ${generator.toString()}\n to\n ${JSON.stringify(context)}`)
          let errorMessage = 'The apply function did not return a value'
          try {
            contextPrime = generator.apply(args, objects, g, args.gs, context, hierarchy, config, response, log)
          } catch( e ) {
            // the next if handle this
            contextPrime = null
            if (e.stack && e.message) {
              errorMessage = `Error applying generator '${generator.notes}'. Error is ${e.toString()} stack is ${e.stack}. Generator is ${generator.toString()}`
            } else if (e.error) {
              errorMessage = `Error applying generator '${generator.notes}'. Error is ${e.error.join()}. Generator is ${generator.toString()}`
            }
          }
          if (!contextPrime && contextPrime !== '') {
            const widths = [10, 10, 90]
            const lines = new Lines(widths)
            lines.setElement(0, 0, 'Generator')
            const source = `${generator.km}/#${generator.index}`
            lines.setElement(0, 2, `ERROR while applying (${source}) ${generator.toLabel()}`)
            lines.newRow()
            lines.setElement(0, 2, generator.toString())
            lines.newRow()
            lines.setElement(0, 1, 'TO')
            lines.setElement(0, 2, `(HASHCODE ${helpers.hashCode(JSON.stringify(sortJson(context, { depth: 25 })))})`)
            lines.setElement(1, 2, JSON.stringify(sortJson(context, { depth: 25 }), null, 2))
            lines.newRow()
            lines.setElement(0, 1, 'STACK')
            lines.setElement(0, 2, stack)
            lines.newRow()
            lines.setElement(0, 1, 'ERROR')
            lines.setElement(0, 2, errorMessage)
            this.logs.push(lines.toString())
            const message = `ERROR while applying (${source}) ${generator.toLabel()}\n to\n ${JSON.stringify(context, null, 2)}.\n${errorMessage}'`
            // this.logs.push(message)
            // return [message]
            args.calls.pop()
            throw { error: [message], logs: this.logs }
          }
          if (((config || {}).config || {}).debug) {
            const widths = [10, 10, 90]
            const lines = new Lines(widths)
            lines.setElement(0, 0, 'Generator')
            if (generator.index > -1 && generator.km) {
              // lines.setElement(0, 2, `KM '${generator.km}' ordinal: ${generator.index}`)
              lines.setElement(0, 2, generator.toLabel())
            }
            lines.newRow()
            lines.setElement(0, 1, 'APPLIED')
            lines.setElement(0, 2, generator.toLabel())
            lines.newRow()
            lines.setElement(0, 2, generator.toString())
            lines.newRow()
            lines.setElement(0, 1, 'RESULT')
            lines.setElement(0, 2, contextPrime)
            lines.newRow()
            lines.setElement(0, 1, 'STACK')
            lines.setElement(0, 2, stack)
            lines.newRow()
            lines.setElement(0, 1, 'TO')
            lines.setElement(0, 2, `(HASHCODE ${helpers.hashCode(JSON.stringify(sortJson(context, { depth: 25 })))})`)
            lines.setElement(1, 2, JSON.stringify(sortJson(context, { depth: 25 }), null, 2))
            this.logs.push(lines.toString())
          }
          applied = true
          break
        }
      }
      args.calls.pop()
      if (!applied && ((config || {}).config || {}).debug) {
        const widths = [10, 10, 90]
        const lines = new Lines(widths)
        lines.setElement(0, 0, 'Generator')
        lines.setElement(0, 2, 'No generator applied')
        lines.newRow()
        lines.setElement(0, 1, 'TO')
        lines.setElement(0, 2, JSON.stringify(context, null, 2))
        this.logs.push(lines.toString())
      }
      contextsPrime.push(contextPrime)
    }
    return contextsPrime
  }
}

module.exports = { Generator, Generators, normalizeGenerator }
